# -*- coding: utf-8 -*-

import os
import json
import time
import random
import urllib
import logging
from uuid import uuid4

from tornado import web
from tornado import gen

from litedfs.name.handlers.base import BaseHandler, BaseSocketHandler, auth_check
from litedfs.name.utils.fs_core import FileSystemTree, InvalidValueError, SameNameExistsError, TargetPathMustDirectoryError, TargetPathNotExistsError, SourcePathNotExistsError, FileNotExistsError, SameNameFileExistsError
from litedfs.name.utils.listener import Connection
from litedfs.name.utils.common import file_sha1sum, file_md5sum, Errors, splitall
from litedfs.name.config import CONFIG

LOG = logging.getLogger("__name__")


class GenerateFileBlockListHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            file_path = self.get_argument("path", "")
            file_size = int(self.get_argument("size", "0"))
            replica = int(self.get_argument("replica", "1"))
            lock_ttl = int(self.get_argument("lock_ttl", "60"))
            block_size = CONFIG["block_size"]
            if file_path:
                fs = FileSystemTree.instance()
                if fs:
                    if not fs.exists(file_path):
                        lock_success = fs.set_file_lock(file_path, ttl = lock_ttl)
                        if lock_success:
                            data_nodes = Connection.get_node_infos(without_full_node = True)
                            for i in data_nodes:
                                data_node = data_nodes[i]
                                if data_node[0] == "127.0.0.1":
                                    host_parts = urllib.parse.urlsplit("//" + self.request.host)
                                    data_node[0] = host_parts.hostname
                            if len(data_nodes) > 0:
                                data_node_ids = list(data_nodes.keys())
                                blocks = []
                                block_id = 0
                                if replica < 1:
                                    replica = 1
                                if replica > len(data_nodes):
                                    replica = len(data_nodes)
                                while file_size > block_size:
                                    blocks.append((block_id, block_size, random.sample(data_node_ids, replica)))
                                    file_size -= block_size
                                    block_id += 1
                                if file_size > 0:
                                    blocks.append((block_id, file_size, random.sample(data_node_ids, replica)))
                                result["data_nodes"] = data_nodes
                                result["blocks"] = blocks
                                result["id"] = str(uuid4())
                            else:
                                Errors.set_result_error("NoUsableDataNode", result)
                        else:
                            Errors.set_result_error("SetFileLockFailed", result)
                    else:
                        Errors.set_result_error("SameNameExists", result)
                else:
                    Errors.set_result_error("ServiceNotReadyYet", result)
            else:
                Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class ListFileLockHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            fs = FileSystemTree.instance()
            if fs:
                result["data"] = fs.list_file_locks()
            else:
                Errors.set_result_error("ServiceNotReadyYet", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class UpdateFileLockHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            file_path = self.get_json_argument("path", "")
            lock_ttl = self.get_json_argument("lock_ttl", 60)
            if file_path and lock_ttl:
                fs = FileSystemTree.instance()
                if fs:
                    success = fs.update_file_lock(file_path, ttl = lock_ttl)
                    if not success:
                        Errors.set_result_error("OperationFailed", result)
                else:
                    Errors.set_result_error("ServiceNotReadyYet", result)
            else:
                Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class CreateFileHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def post(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            file_size = int(self.get_json_argument("size", "0"))
            file_path = self.get_json_argument("path", "")
            file_id = self.get_json_argument("id", "")
            replica = int(self.get_json_argument("replica", "1"))
            blocks = self.get_json_argument("blocks", [])
            checksum = self.get_json_argument("checksum", "")
            current_replica = replica
            for block in blocks:
                if current_replica > len(block[2]):
                    current_replica = len(block[2])
            if file_path and file_id and replica:
                fs = FileSystemTree.instance()
                if fs:
                    if not fs.exists(file_path):
                        now = int(time.time())
                        success = fs.create(
                            file_path,
                            {
                                "size": file_size,
                                "id": file_id,
                                "replica": replica,
                                "current_replica": current_replica,
                                "blocks": blocks,
                                "checksum": checksum,
                                "ctime": now,
                                "mtime": now,
                            }
                        )
                        if success:
                            fs.unset_file_lock(file_path)
                        else:
                            Errors.set_result_error("OperationFailed", result)
                    else:
                        Errors.set_result_error("SameNameExists", result)
                else:
                    Errors.set_result_error("ServiceNotReadyYet", result)
            else:
                Errors.set_result_error("InvalidParameters", result)
        except InvalidValueError as e:
            LOG.error(e)
            Errors.set_result_error("InvalidParameters", result)
        except SameNameExistsError as e:
            LOG.error(e)
            Errors.set_result_error("SameNameExists", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class MoveFileDirectoryHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            source_path = self.get_json_argument("source_path", "")
            target_path = self.get_json_argument("target_path", "")
            if source_path and target_path:
                fs = FileSystemTree.instance()
                if fs:
                    success = fs.move(source_path, target_path)
                    if not success:
                        Errors.set_result_error("OperationFailed", result)
                else:
                    Errors.set_result_error("ServiceNotReadyYet", result)
            else:
                Errors.set_result_error("InvalidParameters", result)
        except InvalidValueError as e:
            LOG.error(e)
            Errors.set_result_error("InvalidParameters", result)
        except SameNameExistsError as e:
            LOG.error(e)
            Errors.set_result_error("SameNameExists", result)
        except TargetPathMustDirectoryError as e:
            LOG.error(e)
            Errors.set_result_error("TargetPathMustDirectory", result)
        except TargetPathNotExistsError as e:
            LOG.error(e)
            Errors.set_result_error("TargetPathNotExists", result)
        except SourcePathNotExistsError as e:
            LOG.error(e)
            Errors.set_result_error("SourcePathNotExists", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class RenameFileDirectoryHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            file_path = self.get_json_argument("path", "")
            new_name = self.get_json_argument("new_name", "")
            if file_path and new_name:
                fs = FileSystemTree.instance()
                if fs:
                    success = fs.rename(file_path, new_name)
                    if not success:
                        Errors.set_result_error("OperationFailed", result)
                else:
                    Errors.set_result_error("ServiceNotReadyYet", result)
            else:
                Errors.set_result_error("InvalidParameters", result)
        except InvalidValueError as e:
            LOG.error(e)
            Errors.set_result_error("InvalidParameters", result)
        except SameNameExistsError as e:
            LOG.error(e)
            Errors.set_result_error("SameNameExists", result)
        except FileNotExistsError as e:
            LOG.error(e)
            Errors.set_result_error("FileNotExists", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class UpdateFileHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def put(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            file_path = self.get_json_argument("path", "")
            replica = self.get_json_argument("replica", 1)
            if replica < 1:
                replica = 1
            if file_path and replica:
                fs = FileSystemTree.instance()
                if fs:
                    success = yield fs.update_replica(file_path, replica)
                    if not success:
                        Errors.set_result_error("OperationFailed", result)
                else:
                    Errors.set_result_error("ServiceNotReadyYet", result)
            else:
                Errors.set_result_error("InvalidParameters", result)
        except InvalidValueError as e:
            LOG.error(e)
            Errors.set_result_error("InvalidParameters", result)
        except FileNotExistsError as e:
            LOG.error(e)
            Errors.set_result_error("FileNotExists", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class DeleteFileHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def delete(self):
        result = {"result": Errors.OK}
        try:
            file_path = self.get_argument("path", "")
            if file_path:
                fs = FileSystemTree.instance()
                if fs:
                    success = fs.delete(file_path)
                    if not success:
                        Errors.set_result_error("OperationFailed", result)
                else:
                    Errors.set_result_error("ServiceNotReadyYet", result)
            else:
                Errors.set_result_error("InvalidParameters", result)
        except InvalidValueError as e:
            LOG.error(e)
            Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class CreateDirectoryHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def post(self):
        result = {"result": Errors.OK}
        try:
            self.json_data = json.loads(self.request.body.decode("utf-8"))
            dir_path = self.get_json_argument("path", "")
            if dir_path:
                fs = FileSystemTree.instance()
                if fs:
                    success = fs.makedirs(dir_path)
                    if not success:
                        Errors.set_result_error("OperationFailed", result)
                else:
                    Errors.set_result_error("ServiceNotReadyYet", result)
            else:
                Errors.set_result_error("InvalidParameters", result)
        except InvalidValueError as e:
            LOG.error(e)
            Errors.set_result_error("InvalidParameters", result)
        except SameNameFileExistsError as e:
            LOG.error(e)
            Errors.set_result_error("SameNameFileExists", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class DeleteDirectoryHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def delete(self):
        result = {"result": Errors.OK}
        try:
            dir_path = self.get_argument("path", "")
            if dir_path:
                fs = FileSystemTree.instance()
                if fs:
                    success = fs.delete(dir_path)
                    if not success:
                        Errors.set_result_error("OperationFailed", result)
                else:
                    Errors.set_result_error("ServiceNotReadyYet", result)
            else:
                Errors.set_result_error("InvalidParameters", result)
        except InvalidValueError as e:
            LOG.error(e)
            Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class ListDirectoryHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            dir_path = self.get_argument("path", "")
            include_file = True if self.get_argument("include_file", "true") == "true" else False
            include_directory = True if self.get_argument("include_directory", "true") == "true" else False
            offset = int(self.get_argument("offset", 0))
            limit = int(self.get_argument("limit", 0))
            if dir_path:
                fs = FileSystemTree.instance()
                if fs:
                    r = fs.list_dir(dir_path, offset = offset, limit = limit, recursive = False, include_file = include_file, include_directory = include_directory)
                    result["children"] = r["files"]
                    result["offset"] = r["offset"]
                    result["limit"] = r["limit"]
                    result["total"] = r["total"]
                else:
                    Errors.set_result_error("ServiceNotReadyYet", result)
            else:
                Errors.set_result_error("InvalidParameters", result)
        except InvalidValueError as e:
            LOG.error(e)
            Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class GetFileBlockInfoHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            file_path = self.get_argument("path", "")
            data_nodes = Connection.get_node_infos()
            for i in data_nodes:
                data_node = data_nodes[i]
                if data_node[0] == "127.0.0.1":
                    host_parts = urllib.parse.urlsplit("//" + self.request.host)
                    data_node[0] = host_parts.hostname
            if file_path:
                fs = FileSystemTree.instance()
                if fs:
                    if fs.isfile(file_path):
                        file_info = fs.get_file_info(file_path)
                        if file_info:
                            result["file_info"] = file_info
                            result["data_nodes"] = data_nodes
                            result["block_size"] = CONFIG["block_size"]
                        else:
                            Errors.set_result_error("FileNotExists", result)
                    else:
                        Errors.set_result_error("InvalidParameters", result)
                else:
                    Errors.set_result_error("ServiceNotReadyYet", result)
            else:
                Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class IsFileHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            file_path = self.get_argument("path", "")
            if file_path:
                fs = FileSystemTree.instance()
                if fs:
                    success = fs.isfile(file_path)
                    if success:
                        result["file"] = True
                    else:
                        result["file"] = False
                else:
                    Errors.set_result_error("ServiceNotReadyYet", result)
            else:
                Errors.set_result_error("InvalidParameters", result)
        except InvalidValueError as e:
            LOG.error(e)
            Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class IsDirectoryHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            file_path = self.get_argument("path", "")
            if file_path:
                fs = FileSystemTree.instance()
                if fs:
                    success = fs.isdir(file_path)
                    if success:
                        result["directory"] = True
                    else:
                        result["directory"] = False
                else:
                    Errors.set_result_error("ServiceNotReadyYet", result)
            else:
                Errors.set_result_error("InvalidParameters", result)
        except InvalidValueError as e:
            LOG.error(e)
            Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()


class PathInfoHandler(BaseHandler):
    @auth_check
    @gen.coroutine
    def get(self):
        result = {"result": Errors.OK}
        try:
            file_path = self.get_argument("path", "")
            if file_path:
                fs = FileSystemTree.instance()
                if fs:
                    r = fs.get_info(file_path)
                    result["info"] = {}
                    result["info"]["exists"] = r[0]
                    result["info"]["type"] = "file"
                    if r[1] == "d":
                        result["info"]["type"] = "directory"
                else:
                    Errors.set_result_error("ServiceNotReadyYet", result)
            else:
                Errors.set_result_error("InvalidParameters", result)
        except InvalidValueError as e:
            LOG.error(e)
            Errors.set_result_error("InvalidParameters", result)
        except Exception as e:
            LOG.exception(e)
            Errors.set_result_error("ServerException", result)
        self.write(result)
        self.finish()



# -*- coding: utf-8 -*-

import os
import sys
import json
import logging

cwd = os.path.split(os.path.realpath(__file__))[0]
sys.path.insert(0, os.path.split(cwd)[0])

from litedfs.name.utils.append_log import AppendLogJson
from litedfs.name import logger

LOG = logging.getLogger(__name__)


if __name__ == "__main__":
    logger.config_logging(file_name = "test_append_log.log",
                          log_level = "DEBUG",
                          dir_name = os.path.join(cwd, "logs"),
                          day_rotate = False,
                          when = "D",
                          interval = 1,
                          max_size = 20,
                          backup_count = 5,
                          console = True)

    LOG.debug("test start")
    
    try:
        a = AppendLogJson("./test.log")
        for i in range(10):
            a.writeline({"line_number": i + 1, "message": "this is a test"})
        a.indexlines()
        LOG.debug(a.lines_pos)
        for line in a.iterlines(start = 1):
            LOG.debug("%s, %s", line, type(line))
        LOG.debug(a.readline(5))
        LOG.debug(a.lines())
    except Exception as e:
        LOG.exception(e)

    LOG.debug("test end")



# -*- coding: utf-8 -*-

import os
import sys
import json
import logging

cwd = os.path.split(os.path.realpath(__file__))[0]
sys.path.insert(0, os.path.split(cwd)[0])

from litedfs.name import logger

LOG = logging.getLogger(__name__)


if __name__ == "__main__":
    logger.config_logging(file_name = "test_blocks.log",
                          log_level = "DEBUG",
                          dir_name = os.path.join(cwd, "logs"),
                          day_rotate = False,
                          when = "D",
                          interval = 1,
                          max_size = 20,
                          backup_count = 5,
                          console = True)

    LOG.debug("test start")
    
    try:
        file_size = 103
        replica = 1
        block_size = 100
        blocks = []
        block_id = 0
        while file_size > block_size:
            blocks.append((block_id, block_size))
            file_size -= block_size
            block_id += 1
        if file_size > 0:
            blocks.append((block_id, file_size))
        LOG.debug("blocks: %s", blocks)
    except Exception as e:
        LOG.exception(e)

    LOG.debug("test end")



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-

import os
import re
import sys
import cmd
import json
import time
import hashlib
import argparse
import logging
import random
from io import BytesIO
from getpass import getpass
from base64 import b64encode, b64decode

import requests
from tea_encrypt import EncryptStr, DecryptStr

from litedfs.version import __version__
from litedfs_client.client import LiteDFSClient


BUF_SIZE = 65536


def print_table_result(data, fields, column_width):
    fields.insert(0, "#")
    field_length_map = {}
    lines = []
    num = 1
    column_max_width = column_width
    for field in fields:
        field_length_map[field] = len(field)
    for item in data:
        line = []
        for field in fields:
            if field == "#":
                line.append(str(num))
            else:
                v = str(item[field]) if field in item else ""
                v_len = len(v)
                if column_max_width > 0 and v_len > column_max_width:
                    v_len = column_max_width
                    v = v[:v_len]
                if v_len > field_length_map[field]:
                    field_length_map[field] = v_len
                line.append(v)
        lines.append(tuple(line))
        num += 1
    field_length_map["#"] = len(str(num))
    format_str = ""
    for field in fields:
        field_len = field_length_map[field]
        if field == "#":
            format_str += "%" + " %s" % field_len + "s | "
        else:
            format_str += "%" + "-%s" % field_len + "s | "
    format_str = format_str[:-3]
    print(format_str % tuple(fields))
    for line in lines:
        print(format_str % line)


def file_md5sum(file_path):
    md5 = hashlib.md5()
    with open(file_path, 'rb') as fp:
        while True:
            data = fp.read(BUF_SIZE)
            if not data:
                break
            md5.update(data)
    return md5.hexdigest()


def bytes_io_md5sum(fp):
    md5 = hashlib.md5()
    while True:
        data = fp.read(BUF_SIZE)
        if not data:
            break
        md5.update(data)
    return md5.hexdigest()


def bytes_md5sum(b):
    md5 = hashlib.md5()
    md5.update(b)
    return md5.hexdigest()


def strings_md5sum(l):
    md5 = hashlib.md5()
    for s in l:
        if s:
            md5.update(s.encode("utf-8"))
    return md5.hexdigest()


class NonExitArgumentParser(argparse.ArgumentParser):
    def __init__(self, *args, **kargs):
        super().__init__(*args, **kargs)
        self.already_print_help = False

    def print_help(self, file = None):
        self.already_print_help = True
        if file is None:
            file = sys.stdout
        self._print_message(self.format_help(), file)

    def exit(self, status = 0, message = None):
        pass

    def error(self, message):
        if self.already_print_help:
            self.already_print_help = False
        else:
            self.print_usage(sys.stderr)


class LDFSShell(cmd.Cmd):
    def __init__(self, host, port, user = None, password = None, column_width = 0):
        super().__init__()
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.column_width = column_width
        self.ldfs = LiteDFSClient(self.host, self.port, user = self.user, password = self.password)
        self.intro = ("LiteDFS Client %s\n" % __version__ +
                      "Connect to Service<%s:%s>\n" % (self.host, self.port) +
                      "Type 'help' or '?' to list commands, Type 'exit' to exit.\n")
        self.prompt = '> '

    def precmd(self, line):
        return line

    def do_about(self, arg):
        "about"
        r = requests.get("http://%s:%s/" % (self.host, self.port))
        print(r.json())

    def do_ls(self, arg):
        "remote list directory: ls -r /path -o 0 -l 10 -f -d"
        try:
            p = NonExitArgumentParser(prog = "ls", add_help = False)
            p.add_argument("-h", "--help", help = "", action = "help")
            p.add_argument("-r", "--remote-path", required = True, help = "remote directory path", default = "")
            p.add_argument("-o", "--offset", help = "list offset", type = int, default = 0)
            p.add_argument("-l", "--limit", help = "list limit", type = int, default = 0)
            p.add_argument("-f", "--exclude-file", help = "exclude file", action = "store_false")
            p.add_argument("-d", "--exclude-directory", help = "exclude directory", action = "store_false")
            args = p.parse_args(arg.split())
            if args.remote_path:
                try:
                    r = self.ldfs.list_directory(args.remote_path, offset = args.offset, limit = args.limit, include_file = args.exclude_file, include_directory = args.exclude_directory)
                    if r:
                        print_table_result(
                            r["children"],
                            ["id", "type", "size", "name"],
                            self.column_width
                        )
                    else:
                        print("list directory[%s] failed" % args.remote_path)
                except Exception as e:
                    print(e)
        except Exception as e:
            print(e)

    def do_mkdir(self, arg):
        "remote create directory: mkdir -r /path"
        try:
            p = NonExitArgumentParser(prog = "mkdir", add_help = False)
            p.add_argument("-h", "--help", help = "", action = "help")
            p.add_argument("-r", "--remote-path", required = True, help = "remote directory path", default = "")
            args = p.parse_args(arg.split())
            if args.remote_path:
                try:
                    r = self.ldfs.create_directory(args.remote_path)
                    if r:
                        print("create directory[%s] success" % args.remote_path)
                    else:
                        print("create directory[%s] failed" % args.remote_path)
                except Exception as e:
                    print(e)
        except Exception as e:
            print(e)

    def do_rm(self, arg):
        "remote delete directory or file: rm -r /path"
        try:
            p = NonExitArgumentParser(prog = "rm", add_help = False)
            p.add_argument("-h", "--help", help = "", action = "help")
            p.add_argument("-r", "--remote-path", required = True, help = "remote directory or file path", default = "")
            args = p.parse_args(arg.split())
            if args.remote_path:
                try:
                    info = self.ldfs.info_path(args.remote_path)
                    if "info" in info and info["info"]["exists"]:
                        if info["info"]["type"] == "directory":
                            r = self.ldfs.delete_directory(args.remote_path)
                            if r:
                                print("delete directory[%s] success" % args.remote_path)
                            else:
                                print("delete directory[%s] failed" % args.remote_path)
                        elif info["info"]["type"] == "file":
                            r = self.ldfs.delete_file(args.remote_path)
                            if r:
                                print("delete file[%s] success" % args.remote_path)
                            else:
                                print("delete file[%s] failed" % args.remote_path)
                except Exception as e:
                    print(e)
        except Exception as e:
            print(e)

    def do_mv(self, arg):
        "remote create directory or file: mv -s /path -t /path"
        try:
            p = NonExitArgumentParser(prog = "mv", add_help = False)
            p.add_argument("-h", "--help", help = "", action = "help")
            p.add_argument("-s", "--source-path", required = True, help = "source directory path", default = "")
            p.add_argument("-t", "--target-path", required = True, help = "target directory path", default = "")
            args = p.parse_args(arg.split())
            if args.source_path and args.target_path:
                try:
                    info = self.ldfs.info_path(args.source_path)
                    if "info" in info and info["info"]["exists"]:
                        if info["info"]["type"] == "directory":
                            r = self.ldfs.move_directory(args.source_path, args.target_path)
                            if r:
                                print("move directory[%s] to %s success" % (args.source_path, args.target_path))
                            else:
                                print("move directory[%s] to %s failed" % (args.source_path, args.target_path))
                        elif info["info"]["type"] == "file":
                            r = self.ldfs.move_file(args.source_path, args.target_path)
                            if r:
                                print("move file[%s] to %s success" % (args.source_path, args.target_path))
                            else:
                                print("move file[%s] to %s failed" % (args.source_path, args.target_path))
                except Exception as e:
                    print(e)
        except Exception as e:
            print(e)

    def do_rename(self, arg):
        "remote create directory or file: rename -r /path -n name"
        try:
            p = NonExitArgumentParser(prog = "rename", add_help = False)
            p.add_argument("-h", "--help", help = "", action = "help")
            p.add_argument("-r", "--remote-path", required = True, help = "remote directory path", default = "")
            p.add_argument("-n", "--new-name", required = True, help = "new directory name", default = "")
            args = p.parse_args(arg.split())
            if args.remote_path and args.new_name:
                try:
                    info = self.ldfs.info_path(args.remote_path)
                    if "info" in info and info["info"]["exists"]:
                        if info["info"]["type"] == "directory":
                            r = self.ldfs.rename_directory(args.remote_path, args.new_name)
                            if r:
                                print("rename directory[%s] to %s success" % (args.remote_path, args.new_name))
                            else:
                                print("rename directory[%s] to %s failed" % (args.remote_path, args.new_name))
                        elif info["info"]["type"] == "file":
                            r = self.ldfs.rename_file(args.remote_path, args.new_name)
                            if r:
                                print("rename file[%s] to %s success" % (args.remote_path, args.new_name))
                            else:
                                print("rename file[%s] to %s failed" % (args.remote_path, args.new_name))
                except Exception as e:
                    print(e)
        except Exception as e:
            print(e)

    def do_upload(self, arg):
        "remote upload file: upload -r /path -l /path -L 120 -R 1"
        try:
            p = NonExitArgumentParser(prog = "upload", add_help = False)
            p.add_argument("-h", "--help", help = "", action = "help")
            p.add_argument("-l", "--local-path", required = True, help = "local file path", default = "")
            p.add_argument("-r", "--remote-path", required = True, help = "remote file path", default = "")
            p.add_argument("-L", "--lock-ttl", help = "lock ttl/seconds, default: 60", type = int, default = 60)
            p.add_argument("-R", "--replica", help = "replica count", type = int, default = 1)
            args = p.parse_args(arg.split())
            try:
                r = self.ldfs.create_file(args.local_path, args.remote_path, replica = args.replica, lock_ttl = args.lock_ttl, progress_callback = print)
                if r:
                    print("create file[%s] success" % args.remote_path)
                else:
                    print("create file[%s] failed" % args.remote_path)
            except Exception as e:
                print(e)
        except Exception as e:
            print(e)

    def do_update(self, arg):
        "remote update file: update -r /path -R 1"
        try:
            p = NonExitArgumentParser(prog = "update", add_help = False)
            p.add_argument("-h", "--help", help = "", action = "help")
            p.add_argument("-r", "--remote-path", required = True, help = "remote file path", default = "")
            p.add_argument("-R", "--replica", required = True, help = "replica count", type = int, default = 1)
            args = p.parse_args(arg.split())
            if args.remote_path and args.replica:
                try:
                    r = self.ldfs.update_file(args.remote_path, args.replica)
                    if r:
                        print("update file[%s] success" % args.remote_path)
                    else:
                        print("update file[%s] failed" % args.remote_path)
                except Exception as e:
                    print(e)
        except Exception as e:
            print(e)

    def do_download(self, arg):
        "remote download file: download -r /path -l /path"
        try:
            p = NonExitArgumentParser(prog = "download", add_help = False)
            p.add_argument("-h", "--help", help = "", action = "help")
            p.add_argument("-l", "--local-path", required = True, help = "local file path", default = "")
            p.add_argument("-r", "--remote-path", required = True, help = "remote file path", default = "")
            args = p.parse_args(arg.split())
            r = self.ldfs.download_file(args.remote_path, args.local_path, progress_callback = print)
            if r:
                print("download file[%s => %s] success" % (args.remote_path, args.local_path))
            else:
                print("download file[%s => %s] failed" % (args.remote_path, args.local_path))
        except Exception as e:
            print(e)

    def do_info(self, arg):
        "remote info file: info -r /path"
        try:
            p = NonExitArgumentParser(prog = "info", add_help = False)
            p.add_argument("-h", "--help", help = "", action = "help")
            p.add_argument("-r", "--remote-path", required = True, help = "remote file path", default = "")
            args = p.parse_args(arg.split())
            if args.remote_path:
                try:
                    r = self.ldfs.info_file(args.remote_path)
                    if r:
                        print(json.dumps(r, indent = 4, sort_keys = True))
                    else:
                        print("get file[%s]'s info failed" % args.remote_path)
                except Exception as e:
                    print(e)
        except Exception as e:
            print(e)

    def do_cluster(self, arg):
        "cluster info: cluster -r"
        try:
            p = NonExitArgumentParser(prog = "cluster", add_help = False)
            p.add_argument("-h", "--help", help = "", action = "help")
            p.add_argument("-r", "--raw", help = "display raw json data", action = "store_true")
            args = p.parse_args(arg.split())
            r = self.ldfs.cluster_info()
            if r:
                if args.raw:
                    print(json.dumps(r, indent = 4, sort_keys = True))
                else:
                    if r["result"] == "ok":
                        print("online nodes:")
                        print_table_result(
                            r["info"]["online_nodes"],
                            [
                                "id",
                                "node_id",
                                "http_host",
                                "http_port",
                                "data_path",
                            ],
                            self.column_width
                        )
                        print("\noffline nodes:")
                        print_table_result(
                            r["info"]["offline_nodes"],
                            [
                                "id",
                                "node_id",
                                "http_host",
                                "http_port",
                                "data_path",
                            ],
                            self.column_width
                        )
                    else:
                        print_table_result(
                            [r],
                            ["result", "message"],
                            self.column_width
                        )
            else:
                print("get cluster info failed")
        except Exception as e:
            print(e)

    def do_path(self, arg):
        "remote path info: path -r /path"
        try:
            p = NonExitArgumentParser(prog = "path", add_help = False)
            p.add_argument("-h", "--help", help = "", action = "help")
            p.add_argument("-r", "--remote-path", required = True, help = "remote file/directory path", default = "")
            args = p.parse_args(arg.split())
            if args.remote_path:
                try:
                    r = self.ldfs.info_path(args.remote_path)
                    if r:
                        print(json.dumps(r, indent = 4, sort_keys = True))
                    else:
                        print("get file/directory[%s]'s info failed" % args.remote_path)
                except Exception as e:
                    print(e)
        except Exception as e:
            print(e)

    def do_exit(self, arg):
        "exit"
        return True


def main():
    parser = argparse.ArgumentParser(prog = 'ldfs', add_help = False)

    # common arguments
    parser.add_argument("address", help = "name node address, host:port")
    parser.add_argument("-w", "--column_width", help = "column max width", type = int, default = 0)
    parser.add_argument("-v", "--version", action = 'version', version = '%(prog)s ' + __version__)
    parser.add_argument("-u", "--user", help = "user name", default = "")
    parser.add_argument("-p", "--password", help = "with password", action = "store_true", default = False)
    args = parser.parse_args()


    try:
        address = args.address
        if address:
            host, port = address.split(":")
            password = None
            if args.password:
                password = getpass("password: ")
            
            shell = LDFSShell(host, port, args.user, password, column_width = args.column_width)
            shell.cmdloop()
    except Exception as e:
        logging.error(logging.traceback.format_exc())


if __name__ == "__main__":
    main()
